# This file is part of CloudControl.
#
# CloudControl is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CloudControl is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with CloudControl.  If not, see <http://www.gnu.org/licenses/>.


"""Plugins helpers for Cloud Control clients."""

from inspect import getmembers

from cloudcontrol.common.client.tags import TagDB


def rpc_handler_decorator_factory(attr_name):
    """Can be used to create custom decorators for marking callable objects as
    RPC handlers

    """
    def rpc_handler(name_or_func=None):
        """RPC handler decorator

        Declares a callable as an RPC handler

        :param name_or_func: allow decorator to be used in two manners:
            # the first case is to export the handler using the same name as the
            # function
            @rpc_handler
            def plop():
                pass

            # the second case allow to specify the RPC handler exported name
            @rpc_handler('plop')
            def plop_handler():
                pass
        """

        def decorator(callable_):
            """Decorator for marking a callable as an RPC handler."""
            assert callable(callable_)
            setattr(callable_, attr_name, name)
            return callable_

        if callable(name_or_func):
            # we are in the first case
            name = True
            return decorator(name_or_func)

        # else, second case
        name = name_or_func
        return decorator
    return rpc_handler


def get_rpc_handlers(o, marker):
    """Get RPC handlers members of an object

    :param o: the object to instrospect
    :param str marker: the attribute to check to decide whether the member is an
        RPC handler
    """
    def get_marker(member):
        return getattr(member, marker, None)

    return dict((
        # return handler name to be the callable object name or the one defined
        # through the decorator
        get_marker(v) if isinstance(get_marker(v), basestring) else k,
        v,
    ) for k, v in getmembers(
        o, lambda m: callable(m) and get_marker(m) is not None))


#: basic decorator to declare RPC handlers
rpc_handler_marker = '_rpc_handler'
rpc_handler = rpc_handler_decorator_factory(rpc_handler_marker)


class Base(object):
    """Example skeleton plugin for clients.

    If you want to create your own plugin, you may create an object that would
    quack just like this one or just inherit from this class.

    """
    def __init__(self, *args, **kwargs):
        """
        :param loop: MainLoop instance
        """
        #: MainLoop instance
        self.main = kwargs.pop('loop')

        # plugins may define tags (see :mod:`ccnode.tags`)
        self.tag_db = TagDB()

        # plugins may define handler functions that would be called by the
        # server
        self.rpc_handler = get_rpc_handlers(self, rpc_handler_marker)

        # tag_db and rpc_handler can be implemented as properties if more logic
        # is needed

    def __hash__(self):
        """This method is used when registering a plugin in the main loop.

        By default, only one instance is allowed. Subclasses can overide this
        method to change this behaviour.

        """
        return hash(self.__class__.__name__)

    def start(self):
        """Used to start pyev watchers."""
        pass

    def stop(self):
        """Cleanup for plugins, can be used to clean pyev watchers."""

        self.main = None
        # TODO dependencies
